/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "para.h"
#include "config.h"
#include "dbtool.cmdline.h"

#include <mysql/mysql.h>
#include <mysql/mysql_version.h>
#include <regex.h>
//#include <libgen.h>
#include <sys/mman.h> /* mmap */


#define MEDIUM_BLOB_SIZE 16777220 /*  (2*24 + 4) */
#define BLOB_SIZE 65539 /* (2*16 + 3) */

static struct gengetopt_args_info args_info;

static int com_vrfy(int argc, char *argv[]);
static int com_us(int argc, char *argv[]);
static int com_upd(int argc, char *argv[]);
static int com_summary(int argc, char *argv[]);
static int com_streams(int argc, char *argv[]);
static int com_stradd_picadd(int argc, char *argv[]);
static int com_strq(int argc, char *argv[]);
static int com_strdel(int argc, char *argv[]);
static int com_sa(int argc, char *argv[]);
static int com_piclist(int argc, char *argv[]);
static int com_picdel(int argc, char *argv[]);
static int com_picch(int argc, char *argv[]);
static int com_picass(int argc, char *argv[]);
static int com_pic(int argc, char *argv[]);
static int com_na(int argc, char *argv[]);
static int com_ls(int argc, char *argv[]);
static int com_last(int argc, char *argv[]);
static int com_laa(int argc, char *argv[]);
static int com_la(int argc, char *argv[]);
static int com_info(int argc, char *argv[]);
static int com_hist(int argc, char *argv[]);
static int com_help(int argc, char *argv[]);
static int com_sl(int argc, char *argv[]);
static int com_da(int argc, char *argv[]);
static int com_cdb(int argc, char *argv[]);
static int com_ca(int argc, char *argv[]);
static int com_verb(int argc, char *argv[]);


/* FIXME: Duplicate */
const struct com {
	char *name;	/* name of the command. */
	unsigned int perms, modus;
	int (*func)();	/* Function to call to do the job */
	char *description; /* one-line descrition */
	char *help;
	char *usage;

} cmds[] = {
	{
		.name = "ca", 
		.func = com_ca, 
		.perms = DB_READ|DB_WRITE, 
		.modus = MODUS_READ,
		.description = "copy attributes",
		.usage = "ca source dest1 [dest2,..]",
		.help = 
		"Copy attributes from source file to destination file(s).\n"
	},
	{
		.name = "cdb", 
		.func = com_cdb, 
		.perms = DB_READ|DB_WRITE, 
		.modus = MODUS_READ,
		.description = "create database",
		.usage = "cdb name",
		.help =
                "Create database name containing the initial columns \n"
                "for basic interoperation with server. This command  \n"
                "has to be used only once when you install this      \n"
                "software for the very first time.                   \n"
	},
	{
		.name = "sl", 
		.func = com_sl, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "print score list",
                .usage = "sl n s",
                .help =
                "Print list of n songs together with their scores    \n"
                "(descending order) in stream s.    \n"
	},
	{ 
		.name = "help", 
		.func = com_help, 
		.perms = 0, 
		.modus = MODUS_READ,
		.description = "", /* help for help is defined in command.c */
		.usage = "",
		.help = ""
	}, 
	{
		.name = "hist", 
		.func = com_hist, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "print history",
		.usage = "hist",
		.help =
                "Print list of all songs together with number of     \n"
                "days since song was last played.                    \n"
	},
	{
		.name = "info", 
		.func = com_info, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "print database info",
		.usage = "info [song]",
		.help = 
                "print database informations for song. Current song  \n"
                "is used if song is not given.                     \n"
	},
	{
		.name = "la", 
		.func = com_la, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "list attributes",
		.usage = "la [name]",
		.help =
                "List attributes of \"name\" or of current song when   \n"
                "invoked without arguments.                          \n"
	},
	{
		.name = "laa", 
		.func = com_laa, 
		.perms = DB_READ,
		.modus = MODUS_READ,
		.description = "list available attributes",
		.usage = "laa",
		.help = "What should I say more?\n"
		
	},
	{
		.name = "last", 
		.func = com_last, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "print list of songs, order by lastplayed",
		.usage = "last [n]",
		.help =
		"The optional number n defaults to 10 if not specified.\n"
	}, 
	{
		.name = "ls",
		.func = com_ls, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "list songs matching a LIKE pattern",
		.usage = "ls [pattern]",
		.help =
		"See the documentation of mysql for the definition of \n"
		"LIKE patterns.\n"
	},
	{
		.name = "sa", 
		.func = com_sa, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "set/unset attributes",
		.usage = "sa attr1{+|-} [attr2{+|-} ...] [name1 ...]",
		.help
                "Set (+) or unset (-) attributes \"attr1, ...\" for    \n"
                "\"name1, ...,\" or current song if no name is given.  \n"
	},
	{
		.name = "skip", 
		.func = com_sl, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "skip subsequent song(s)",
                .usage = "skip n stream",
                .help =
                "Skip \"n\" songs in stream \"stream\", This is achieved \n"
                "by setting the lastplayed time of the subsequent    \n"
                "songs to the current time.                          \n"
	},
	{
		.name = "streams", 
		.func = com_streams, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "",
		.usage = "", /* streams is defined in command.c */
		.help = ""
	},
	{
		.name = "summary", 
		.func = com_summary, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "list attributes",
		.usage = "summary",
		.help
                "Print a list of attributes together with number of  \n"
                "songs having that attribute set.                    \n"
	},
	{
		.name = "us", 
		.func = com_us, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "update lastplayed time",
                .usage = "us name",
                .help =
                "Invoked by server just before song is streamed.    \n"
                "Usually, the user does not need this command at all.\n"
	},
	{
		.name = "na", 
		.func = com_na, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "add new attribute to database",
                .usage = "na name",
                .help =
                "This adds a column named \"name\" to your mysql       \n"
                "database. \"name\" should only contain letters and    \n"
                "numbers, in paricular, '+' and '-' are not allowed  \n"
	},
	{
		.name = "da", 
		.func = com_da, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description  = "drop attribute from database",
		.usage = "da name",
		.help = 
		"Use with caution. All info on attribute \"name\" will \n"
		"be lost.                                            \n"
	},
	{
		.name = "vrfy", 
		.func = com_vrfy, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "list invalid entries in database",
                .usage = "vrfy",
                .help =
                "Print what \"clean\" will change. Should always be    \n"
                "run after updating the database with \"upd\".         \n"
	},
	{
		.name = "clean", 
		.func = com_vrfy, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "nuke invalid entries in database",
                .usage = "clean",
                .help =
                "If \"vrfy\" shows that you have invalid entries in    \n"
                "your database, you can get rid of them with this    \n"
                "command.  It also sets undefined picture IDs to     \n"
                "the default value 1 and fixes NULL entries for      \n"
                "attributes. Always run \"upd\" and \"vrfy\" before      \n"
                "running this command. Use with caution!             \n"
	},
	{
		.name = "upd", 
		.func = com_upd, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "update database",
                .usage = "upd",
                .help =
                "This command uses the find_cmd which was specified  \n"
                "in dbtool.conf to find your sound files. New files  \n"
                "are then added to the mysql database. Use this      \n"
                "command if you got new sound files or if you have   \n"
                "moved some files around.                            \n"
	},
	{
		.name = "pic", 
		.func = com_pic, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "get picture by name or by identifier",
                .usage = "pic [name]",
                .help =
                "Dump jpg image that is associated to given song     \n"
                "(current song if not specified) to stdout. If name  \n"
                "starts with '#' it is interpreted as an identifier  \n"
                "instead and the picture having that identifier is   \n"
                "dumped to stdout.                                   \n"
	},
	{
		.name = "picadd", 
		.func = com_stradd_picadd, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_WRITE,
		.description = "add picture to database",
		.usage = "picadd [name]",
		.help =
                "Read jpeg file from stdin and store it as \"name\" \n"
		"in database.\n"
	},
	{
		.name = "piclist", 
		.func = com_piclist, 
		.perms = DB_READ, 
		.modus = MODUS_READ,
		.description = "print list of pictures",
                .usage = "piclist",
                .help =
                "Print id, name and length of each picture contained \n"
                "in the database.                                    \n"
	},
	{
		.name = "picass", 
		.func = com_picass, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "associate a picture to songs(s)",
                .usage = "picass pic_id song1 [song2...]",
                .help =
                "Associate the picture given by \"id\" to all given songs. \n"
	},
	{
		.name = "picdel", 
		.func = com_picdel, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "delete picture from database",
                .usage = "picdel id1 [id2...]",
                .help =
                "Delete each given picture from database.\n"
	},
	{
		.name = "picch", 
		.func = com_picch, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "change name of picture",
                .usage = "picch id new_name",
                .help =
                "Asign \"new_name\" to picture with identifier \"id\".\n"
	},
	{
		.name = "verb", 
		.func = com_verb, 
		.perms = DB_READ | DB_WRITE, 
		.modus = MODUS_READ,
		.description = "send verbatim sql query",
                .usage = "verb cmd",
                .help =
                "Send cmd to mysql server. For expert/debugging only \n"
                "Note that cmd usually must be escaped.  Use only if \n"
                "you know what you are doing!                        \n"
	},
	{
		.name = "stradd",
		.func = com_stradd_picadd,
		.perms = DB_READ | DB_WRITE,
		.modus = MODUS_WRITE,
		.description = "add stream",
		.usage = "stradd name",
		.help =
                "Add stream name to list of available streams.  The  \n"
                "stream definition is read from stdin and then sent  \n"
                "to the server.                                      \n"
	},
	{
		.name = "strq",
		.func = com_strq,
		.perms = DB_READ,
		.modus = MODUS_READ,
		.description = "query stream definition",
		.usage = "strq name",
		.help =
		"Print definition of stream to stdout.\n"
	},
	{
		.name = "strdel",
		.func = com_strdel,
		.perms = DB_READ | DB_WRITE,
		.modus = MODUS_READ,
		.description = "delete stream",
		.usage = "strdel name",
		.help = 
		"Remove stream from database.\n"
	},
	{
		.name = "", 
		.func = NULL, 
		.perms = 0, 
		.description = "", 
		.help = ""
	} 
};

struct regex_list {
	char *regex;
	char *replacement;
};

/*
 * read options from cmdline and config file. Returns -1 on file errors,
 * 1 on success.
 */
static int read_config(const int argc, char **argv) 
{
	struct stat statbuf;
	char *home;
	char config_file[MAXLINE];

	cmdline_parser(argc, argv, &args_info);
	if (!(home = getenv("HOME")))
		return -1;
	sprintf(config_file, "%s/.paraslash/dbtool.conf", home);
	if (stat(config_file, &statbuf)) {
		printf("Error: Can't stat config file %s\n", config_file);
		return -1;
	}
	cmdline_parser_configfile(config_file, &args_info, 0, 0, 0);
	if (!args_info.mysql_passwd_given) {
		printf("Error: No password specified\n");
		cmdline_parser_print_help();
		return -1;
	}
	if (!args_info.find_cmd_given) {
		printf("Error: No find command specified\n");
		return -1;
	}
	return 1;
}


/*
 * connect to mysql server, return mysql pointer on success, NULL on errors
 */
static struct MYSQL *init() 
{
	void *mysql_ptr;
	struct gengetopt_args_info *a = &args_info;
	mysql_ptr = mysql_init(NULL);

	if (!mysql_ptr) {
		printf("mysql init error\n");
		return NULL;
	}

	if (!(mysql_ptr = mysql_real_connect(mysql_ptr, a->mysql_host_arg, 
			a->mysql_user_arg, a->mysql_passwd_arg, 
			a->mysql_database_arg? a->mysql_database_arg : NULL, 
			a->mysql_port_arg, NULL, 0))) {
		printf("connect error\n");
		return NULL;
	}
	return mysql_ptr;
}

/*
 * Use open connection given by mysql_ptr to query server. Returns a
 * result pointer on succes and NULL on errors
 */
static struct MYSQL_RES *get_result(void *mysql_ptr, char *query) 
{
	int query_ret;
	void *result;

//	printf("q: %s (%d)\n", query, strlen(query));
	query_ret = mysql_real_query(mysql_ptr, query, strlen(query));
	if (query_ret) {
		printf("real_query error (%s)\n", mysql_error(mysql_ptr));
		return NULL;
	}
	result = mysql_store_result(mysql_ptr);
	if (!result) {
		printf("store_result error\n");
		return NULL;
	}
	return result;
}

/* 
 * get all attributes 
 */
static void *getaa(void *mysql_ptr) 
{
	void *result;
	char buf[MAXLINE];

	sprintf(buf, "desc data");
	result = get_result(mysql_ptr, buf);
	if (!result || !mysql_num_rows(result))
		return NULL;
	/* skip first four rows*/
	if (!mysql_fetch_row(result) || !mysql_fetch_row(result) 
			|| !mysql_fetch_row(result) 
			|| !mysql_fetch_row(result)) 
		return NULL;
	return result;
}

/* 
 * print next num results of result to stdout 
 */
static int print_results(void *result, unsigned int top, unsigned int left, 
		unsigned int bottom, unsigned int right) 
{
	unsigned int i,j;
	MYSQL_ROW row;

	for (i = top; i <= bottom; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0]) 
			return -1;
		for (j = left; j <= right; j++)
			printf(j == left? "%s" : "\t%s", row[j]? 
				row[j] : "NULL");
		printf("\n");
	}
	return 0;
}

/*
 * Return number of times song has been played.
 */
static unsigned int get_numplayed(void *mysql_ptr, char *name)
{
	void *result = NULL;
	char buf[STRINGSIZE];
	MYSQL_ROW row;
	int ret;

	sprintf(buf, "select Numplayed from data where name = '%s'", name);
        result = get_result(mysql_ptr, buf);
	ret = 0;
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0]) 
		goto out;
//	printf("row[0]: %s\n", row[0]);
	ret = atoi(row[0]);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

static char *escape_str(void *mysql_ptr, char* old)
{
	char *new = (char*)malloc(2 * strlen(old) * sizeof(char) + 1);

	if (!new) {
		printf("malloc error\n");
		exit(EXIT_FAILURE);
	}
	mysql_real_escape_string(mysql_ptr, new, old, strlen(old));
	return new;
}

static int do_update_song(void *mysql_ptr, char *name)
{
	int ret;
	char buf[STRINGSIZE];
	unsigned int numplayed;
	char *escaped_name = escape_str(mysql_ptr, basename(name));

	sprintf(buf, "update data set lastplayed = now() where "
		"name='%s'", escaped_name);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret)
		goto out;
	numplayed = get_numplayed(mysql_ptr, escaped_name);
//	printf("numplayed: %i\n", numplayed);
	sprintf(buf, "update data set Numplayed=%i where name='%s'", 
		numplayed + 1, escaped_name);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
out:
	if (escaped_name)
		free(escaped_name);
	return ret;
}


/* 
 * print all attributes to stdout 
 */
static int com_laa(int argc, char *argv[]) 
{
	void *result;
	int ret;
	void *mysql_ptr;

	if (!(mysql_ptr = init()))
		return -1;
	result = getaa(mysql_ptr);
	if (!result)
		goto close;
	ret = print_results(result, 0, 0, mysql_num_rows(result) - 1, 0);
	mysql_free_result(result);
close:
	mysql_close(mysql_ptr);
	return ret;
}

/* 
 * get last num songs. mysql connection must already be established 
 */
static void *get_last(void* mysql_ptr, unsigned int num) 
{
	char buf[STRINGSIZE];
	sprintf(buf, "select concat(dir.dir,'/',dir.name) "
		"from data,dir where dir.name=data.name and "
		"to_days(data.lastplayed)!=0 order by lastplayed DESC "
		"limit %u", num);
	return get_result(mysql_ptr, buf);
}

static char *get_very_last(void* mysql_ptr)
{
	void *result = NULL;
	char *ret, *tmp;
	MYSQL_ROW row;

	ret = NULL;
	result = get_last(mysql_ptr, 1);
	if (!result)
		return NULL;
	row = mysql_fetch_row(result); 
	if (!row || !row[0])
		goto out;
	tmp = strdup(row[0]);
	ret = strdup(basename(tmp));
	free(tmp);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* 
 * print songs sorted by lastplayed 
 */
static int com_last(int argc, char *argv[]) 
{
	void *result = NULL;
	void *mysql_ptr;
	MYSQL_ROW row;
	int ret;
	unsigned int num_rows, num;

	if (argc < 3) 
		num = 10;
	else
		num = atoi(argv[2]);
	if (!num)
		return 0;

	if (!(mysql_ptr = init()))
		return -1;
	result = get_last(mysql_ptr, num);
	if (!result) {
		ret = -1;
		goto out;
	}
	num_rows = mysql_num_rows(result);

	while (num_rows-- >=0 ) { 
		mysql_data_seek(result, num_rows);
		row = mysql_fetch_row(result); 
		if (!row || !row[0]) {
			ret = -1;
			goto out;
		}
		printf("%s\n", row[0]);
	}
out:	
	if (result)
		mysql_free_result(result);
	mysql_close(mysql_ptr);
	return ret;
}

/*
 * Simple search and replace routine
 */
static int s_a_r(char *src, char *dest, char* regex, char *replacement)
{
	regex_t preg;
	size_t  nmatch = 1;
	regmatch_t pmatch[1];
	int eflags = 0;
	int ret = 0;
	char *bufptr = src;

//	printf("Checking for %s\n", regex);
	*dest = '\0';
	regcomp(&preg, regex, 0);
	while (regexec(&preg,  bufptr, nmatch, pmatch, eflags) 
			!= REG_NOMATCH) {
		char tmp[STRINGSIZE], arg[STRINGSIZE];
		char *o_bracket;
		char *c_bracket;
		int max;
	
		o_bracket = strchr(bufptr + pmatch[0].rm_so, '(');
		c_bracket = o_bracket? strchr(o_bracket, ')') : NULL;
		ret++;
		if (!c_bracket) {
			printf("Syntax error\n");
			return -1;
		}
		max = STRINGSIZE - 1 - strlen(dest);
		strncat(dest, bufptr, 
			pmatch[0].rm_so < max? pmatch[0].rm_so: max);
		strncpy(arg, o_bracket + 1, c_bracket - o_bracket - 1);
		arg[c_bracket - o_bracket - 1] = '\0';
		snprintf(tmp, STRINGSIZE -1, replacement, arg);
		
		max = STRINGSIZE - 1 - strlen(dest);
		strncat(dest, tmp, max);
		bufptr = c_bracket;
		bufptr++;
	}
	strncat(dest, bufptr, STRINGSIZE - strlen(dest) - 1);
	return ret;
}

/*
 * search src and replace to dest according to each entry in regex_list 
 */
static int s_a_r_array(char *src, char *dest)
{
	char *tmp = strdup(src);
	int ret = 0;
	struct regex_list regex_list[] = { 
		{	.regex = "IS_N_SET", 
			.replacement = "(data.%s != '1')"
		}, {
			.regex = "IS_SET", 
			.replacement = "(data.%s = '1')"
		}, {
			.regex = "PICID", 
			.replacement = "data.Pic_Id"
		}, {
			.regex = "NAME_LIKE", 
			.replacement = "(dir.name like '%s')"
		}, {
			.regex = "DIR_LIKE", 
			.replacement = "(dir.dir like '%s')"
		}, {
			.regex = "LASTPLAYED", 
			.replacement = "FLOOR((UNIX_TIMESTAMP(now())"
				"-UNIX_TIMESTAMP(data.Lastplayed))/60)"
		}, {
			.regex = "NUMPLAYED", 
			.replacement = "data.Numplayed"
		}, {
			.regex = NULL, 
			.replacement = NULL
		}
	};
	struct regex_list *regex_ptr = regex_list; 

	while (tmp && regex_ptr->regex && regex_ptr->replacement) {
		int sar_ret;

		sar_ret = s_a_r(tmp, dest, regex_ptr->regex, 
			regex_ptr->replacement);
		free(tmp);
		if (sar_ret < 0)
			return -1;
		ret += sar_ret;
		tmp = strdup(dest);
		regex_ptr++;
	}
	free(tmp);
	return ret;
}

/*
 * Read stream definition of stream streamname and construct mysql
 * query. mysql connection must already be established
 */
static int get_query(void* mysql_ptr, char *streamname, char *query)
{
	char accept_opts[STRINGSIZE] = "";
	char deny_opts[STRINGSIZE] = "";
	char score[STRINGSIZE] = "";
	char order[STRINGSIZE] = "";
	char select_clause[STRINGSIZE] = "";
	char where_clause[STRINGSIZE] = "";
	char command[STRINGSIZE] = "";
	char tmp[STRINGSIZE] = "";
	int ret = 0;
	void *result;
	MYSQL_ROW row;
	unsigned long *length;
	char *end;

	sprintf(query, "select def from streams where name='%s'", streamname);
	result = get_result(mysql_ptr, query);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	length = mysql_fetch_lengths(result);
//	printf("row: %lu, row[0]: %lx length[0]: %lx\n", row, row[0], 
//		length[0]);
	if (!row || !row[0])
		goto out;
	if (length[0]) {
		end = row[0];
		while (*end) {
			int n;
			char *arg, *line = end;
			
			if (!(end = memchr(line, '\n', length[0])))
				break;
			*end = '\0';
			end++;	
			if (sscanf(line, "%200s%n", command, &n) < 1) 
				continue;
			arg = line + n;
			if (!strcmp(command, "accept:")) {
				char tmp[STRINGSIZE];
				char tmp2[STRINGSIZE];
				
				s_a_r_array(arg, tmp2);
				if (strlen(accept_opts)) {
					sprintf(tmp, " or %s", tmp2);
					strcat(accept_opts, tmp);
				} else
					strcpy(accept_opts, tmp2);
			} else if (!strcmp(command, "deny:")) {
				char tmp[STRINGSIZE];
				char tmp2[STRINGSIZE];
				
				s_a_r_array(arg, tmp2);
				if (strlen(deny_opts)) {
					sprintf(tmp, " or %s", tmp2);
					strcat(deny_opts, tmp);
				} else
					strcpy(deny_opts, tmp2);
			} else if (!strcmp(command, "score:"))
				s_a_r_array(arg, score);
		}
	}
	if (!*score)
		s_a_r_array(args_info.default_score_arg, score);
	sprintf(order, "order by -(%s)", score);
	sprintf(tmp, "FLOOR(%s)", score);
	strcpy(score, tmp);
	sprintf(select_clause, "select %s, concat(dir.dir, '/', data.name) "
		"from data,dir where dir.name = data.name ", score);
	if (strlen(accept_opts) || strlen(deny_opts)) {
		if (!strlen(accept_opts))
			sprintf(where_clause, "and not (%s) ", deny_opts);
		else {
			if (!strlen(deny_opts))
				sprintf(where_clause, "and (%s) ", accept_opts);
			else
				sprintf(where_clause, 
					"and ((%s) and not (%s)) ", 
					accept_opts, deny_opts);
		}	
	}
	sprintf(query, "%s %s %s", select_clause, where_clause, order);
	ret = 0;
out:
	return ret;
}

/*
 * sl/skip
 */
static int com_sl(int argc, char *argv[]) 
{
	void *mysql_ptr;
	void *result = NULL;
	MYSQL_ROW row;
	int ret, i, skip = 0;
	char query[STRINGSIZE];
	char *streamname = NULL;
	unsigned int num_rows, num;

	if (argc < 4) {
		printf("Syntax error\n");
		return -1;
	}
	num = atoi(argv[2]);
	if (!num)
		return -1;
	streamname = argv[3];
	if (!strcmp(argv[1], "skip"))
		skip = 1;
	
	if (!(mysql_ptr = init()))
		return -1;
	if (get_query(mysql_ptr, streamname, query) < 0)
		goto out;
//	printf("%s\n", query);
	result = get_result(mysql_ptr, query);
	if (!result) {
		printf("get_result error\n");
		ret = -1;
		goto out;
	}
	num_rows = mysql_num_rows(result);
	if (!num_rows) {
		ret = -1;
		goto out;
	}
	for (i = 0; i < num_rows && i < num; i++) {
		row = mysql_fetch_row(result); 
		if (skip) { 
			printf("Skipping %s\n", row[1]);
			do_update_song(mysql_ptr, row[1]);
		} else 
			printf("%s %s\n", row[0]? row[0]: "0", row[1]);
	}
	ret = 0;
out:
	if (result)
		mysql_free_result(result);
	mysql_close(mysql_ptr);
	return ret;
}

/*
 * ls
 */
static int com_ls(int argc, char *argv[]) 
{
	char buf[STRINGSIZE];
	void *mysql_ptr;
	void *result;
	int ret;
	unsigned int num_rows;
	if (argc >= 3) 
		sprintf(buf,"select name from data where name LIKE '%s'",
			argv[2]);
	else
		sprintf(buf,"select name from data");
	if (!(mysql_ptr = init()))
		return -1;
        result = get_result(mysql_ptr, buf);
	if (!result) {
		printf("mysql get_result error\n");
		ret = -1;
		goto close;
	}
	num_rows = mysql_num_rows(result);
//	fprintf(stderr,"got %i rows\n",num_rows);
	if (!num_rows) {
		ret = 0;
		goto free_close;
	}
	ret = print_results(result, 0, 0, num_rows, 0);

free_close:	
	mysql_free_result(result);
close:	
	mysql_close(mysql_ptr);
	return ret;
}

/*
 * help
 */
static int com_help(int argc, char *argv[]) 
{
	int i;

	if (argc < 3) {
		for (i = 0; cmds[i].func; i++) {
			/* skip help and streams command */
			if (!strcmp(cmds[i].name, "streams") || 
				!strcmp(cmds[i].name, "help"))
				continue;
			printf("%s\t%u\t%u\n", cmds[i].name, cmds[i].perms,
				cmds[i].modus); 
		}
		return 0;
	}
	for (i = 0; cmds[i].func; i++) {
		if (!strcmp(cmds[i].name, argv[2])) {
			printf( "%s\n%s\n%s",
				cmds[i].description,
				cmds[i].usage,
				cmds[i].help
			);
			return 0;
		}
	}
	printf("No such command: %s\n", argv[2]);
	return -1;
}

/*
 * update song
 */
static int com_us(int argc, char *argv[]) 
{
	void *mysql_ptr;
	int ret;

	if (argc < 3) 
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	ret = do_update_song(mysql_ptr, argv[2]);
	mysql_close(mysql_ptr);
	return ret;
}

/* get attributes by name. If verbose is not 0, get_a writes a string
 * into atts of the form 'att1="0",att2="1"', which is used in com_ca
 * for contructing a mysql update query.
 */
static int get_a(void *mysql_ptr, char *name, char *atts, int verbose) 
{
	void *result2 = NULL;
	void *result = NULL;
	MYSQL_ROW row, row2;
	char csv[STRINGSIZE] = "";
	char buf[STRINGSIZE];
	int i, ret;
	unsigned int num_fields;

	result2 = getaa(mysql_ptr);
	if (!result2) {
		ret = -1;
		goto out;
	}
	while((row = mysql_fetch_row(result2)) && row && row[0]) {
		if (csv[0])
			strcat(csv, ",");
		strcat(csv, row[0]);
	}
//	printf("all attributes: %s\n",csv);
	sprintf(buf, "select %s from data where name='%s'", csv, name);
//	printf("query: %s\n",buf);
	result = get_result(mysql_ptr, buf);
	if (!result) {
		ret = -1;
		goto out;
	}
	num_fields = mysql_num_fields(result);
//	printf("num_fields: %u\n",num_fields);
	/* skip Lastplayed and Numplayed */
        mysql_data_seek(result2, 4);
	row = mysql_fetch_row(result);
	if (!row) {
		ret = -1;
		goto out;
	}
	*atts = '\0';
	for (i = 0; i < num_fields; i++) {
		row2 = mysql_fetch_row(result2);

                if (row[i] && !strcmp(row[i], "1"))  {
			if (verbose) {
				if (atts[0])
					strcat(atts, ",");
				strcat(atts, row2[0]);
				strcat(atts, "=\"1\"");
			} else {
				if (atts[0])
					strcat(atts, " ");
				strcat(atts, row2[0]);
			}
		} else {
			if (verbose) {
				if (atts[0])
					strcat(atts, ",");
				strcat(atts, row2[0]);
				strcat(atts, "=\"0\"");
			}	
		}
        }
	ret = 0;
out:	
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	return ret;
}

/* 
 * update attributes of name 
 */
static int update_atts(void *mysql_ptr, char *name, char *query) 
{
	char buf[STRINGSIZE];
	int ret;

//	printf("query: %s\n", query);
	if (get_a(mysql_ptr, name, buf, 0) >= 0)
		printf("old: %s\n", buf);
	else {
		printf("update_atts: get attributes failed\n");
		return -1;
	}
	ret = mysql_real_query(mysql_ptr, query, strlen(query));
	if (ret) {
		printf("update_atts: mysql_real_query error\n");
		return ret;
	}
	if (get_a(mysql_ptr, name, buf, 0) >= 0)
		printf("new: %s\n", buf);
	return ret;
}

/*
 * copy attributes
 */
static int com_ca(int argc, char *argv[]) 
{
	char *name = NULL;
	void *mysql_ptr = NULL;
	int i, ret = -1;
	char buf[STRINGSIZE];
	char buf2[STRINGSIZE];

	if (argc < 4) {
		printf("Syntax Error\n");
		return -1;
	}
	if (!(mysql_ptr = init()))
		return -1;
	name = escape_str(mysql_ptr, basename(argv[2]));
	if (!name)
		goto out;
	ret = get_a(mysql_ptr, name, buf, 1);
	free(name);
	if (ret < 0) {
		printf("Get attributes failed\n");
		goto out;
	}
	for (i = 3; i < argc; i++) {
		name = escape_str(mysql_ptr, basename(argv[i]));
		if (!name)
			goto out;
		snprintf(buf2, STRINGSIZE - 1, "update data set %s where "
			"name = '%s'", buf, name);
		//printf("Query: %s\n", buf2);
		printf("Updating %s\n", name);
		update_atts(mysql_ptr, name, buf2);
		free(name);
	}
	ret = 0;
out:
	mysql_close(mysql_ptr);
	return ret;
}

/*
 * list attributes
 */
static int com_la(int argc, char *argv[]) 
{ 
	char *name, *name2;
	void *mysql_ptr = NULL;
	int ret;
	void *result = NULL;
	char buf[STRINGSIZE];
	 
	if (!(mysql_ptr = init()))
		return -1;
	if (argc < 3) {
		name = get_very_last(mysql_ptr);
		if (!name)
			goto out;
		printf("%s\n", name);
	} else 
		name = basename(argv[2]);
	name2 = escape_str(mysql_ptr, name);
	name  = name2;
	ret = get_a(mysql_ptr, name, buf, 0);
	if (ret >= 0)
		printf("%s\n", buf);
	else
		printf("%s: Not contained in database\n", name);

out:	
	if (name)
		free(name);
	if (result)
		mysql_free_result(result);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * verbatim
 */
static int com_verb(int argc, char *argv[]) 
{ 
	void *mysql_ptr = NULL;
	void *result = NULL;
	int ret = -1;
	unsigned int num_rows, num_fields;
	 
	if (argc < 3) {
		printf("verb: Argument required\n");
		return -1;
	}
	if (!(mysql_ptr = init()))
		return -1;
        result = get_result(mysql_ptr, argv[2]);
	if (!result) {
		printf("verb: get_result failed\n");
		return -1;
	}
	num_fields = mysql_field_count(mysql_ptr);
	num_rows = mysql_num_rows(result);
	if (!num_fields || !num_rows) {
		printf("verb: no results\n");
		goto out;
	}
	printf("verb: Found %d rows, %d fields\n", num_rows, num_fields);
	ret = print_results(result, 0, 0, num_rows - 1, num_fields - 1);
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * history
 */
static int com_hist(int argc, char *argv[]) { 
	void *mysql_ptr;
	int ret;
	void *result = NULL;
	char buf[STRINGSIZE];
	unsigned int num_rows;

	if (!(mysql_ptr = init()))
		return -1;
	sprintf(buf, "select name,to_days(now())-to_days(lastplayed) from "
		"data where %s%s lastplayed order by lastplayed",
		(argc < 3)? "" : argv[2] , (argc<3)? "" : "='1' and" );
//	printf("query: %s\n",buf);
        result = get_result(mysql_ptr, buf);
	if (!result) {
		ret = -1;
		goto out;
	}
	num_rows = mysql_num_rows(result);
	print_results(result, 0, 0, num_rows - 1, 1);
	ret = 0;
out:	
	if (result)
		mysql_free_result(result);
	mysql_close(mysql_ptr);
	return ret;
}
#if 0
/* 
 * get dir of given song. song should be a basename 
 */
static int get_dir(void *mysql_ptr, char *song, char *dir) 
{
	char buf[STRINGSIZE];
	void *result = NULL;
	MYSQL_ROW row;
	int ret;

	snprintf(buf, STRINGSIZE - 1, "select dir "
		"from dir where dir.name = '%s'", song);
//	printf("query: %s\n",buf);
        result = get_result(mysql_ptr, buf);
	if (!result) {
		ret = -1;
		goto out;
	}
	row = mysql_fetch_row(result);
	if (!row || !row[0]) {
		ret = -1;
		goto out;
	}
	strcpy(dir, row[0]);
	ret = 0;
out:	
	if (result)
		mysql_free_result(result);
	return ret;
}
#endif

/*
 * info
 */
static int com_info(int argc, char *argv[]) 
{ 
	int ret;
	MYSQL_ROW row;
	char *name, *esc_name;
	void *mysql_ptr = NULL;
	void *result = NULL;
	char buf[STRINGSIZE];
	
	ret = -1;
	if (!(mysql_ptr = init()))
		return -1;
	if (argc < 3) {
		if (!(name = get_very_last(mysql_ptr)))
			goto out;
		printf("%s\n", name);
	} else
		name = strdup(basename(argv[2]));
	esc_name = escape_str(mysql_ptr, name);
	if (!esc_name)
		goto out;
	snprintf(buf, STRINGSIZE - 1,
		"select dir.dir from data, dir where "
		"dir.name = data.name and data.name = '%s'", esc_name);
//	printf("Query: %s\n", buf);
	ret = -1;
        result = get_result(mysql_ptr, buf);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row) {
		printf("Not contained in database\n");
		goto out;
	}
	if (!row[0]) 
		goto out;
	printf("dir: %s\n", row[0]);
	mysql_free_result(result);
	snprintf(buf, STRINGSIZE - 1,
		"select concat('last played: ', "
		"(to_days(now()) - to_days(data.lastplayed)),"
		"' day(s) ago. Numplayed: ',data.Numplayed, "
		"', Pic_Id: ', data.Pic_Id) " 
		"from data,dir where dir.name = data.name and data.name = '%s'", 
		esc_name);
//	printf("Query: %s\n", buf);
        result = get_result(mysql_ptr, buf);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row) { 
		printf("Not contained in database\n");
		goto out;
	}
	if (row[0]) 
		printf("%s\n", row[0]);
	else 
		printf("Not yet played\n");
	ret = get_a(mysql_ptr, esc_name, buf, 0);
	printf("attributes: %s\n", (ret >= 0 && buf[0])?
		buf: "(none)");
out:	
	if (result)
		mysql_free_result(result);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	if (name)
		free(name);
	if (esc_name)
		free(esc_name);
	return ret;
}

/*
 * set attributes
 */
static int com_sa(int argc, char *argv[]) 
{
	int i;
	char buf[STRINGSIZE] = "";
	char buf2[STRINGSIZE] = "";
	int ret;
	void *mysql_ptr = NULL;
	void *result = NULL;
	char *bn; /* basename */

	if (argc < 3) {
		printf("sa: Syntax error\n");
		return -1;
	}
	i = 2;
	while (i < argc) {
		int unset = 0;
		char *p =argv[i];
		int len = strlen(p);
 
		switch (p[len - 1]) {
			case '+':
				unset = 0;
				break;
			case '-':
				unset = 1;
				break;
			default:
				goto no_more_atts;
		}
		p[len - 1] = '\0';
		sprintf(buf2, "%s%s='%s'", buf[0]? "," : "", p, 
			unset? "0" : "1" );
		strcat(buf, buf2);
		i++;
	}
no_more_atts:
	if (!buf[0])
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	if (i >= argc) {
		char *tmp;
		/* no name given, use current song */
		if (!(bn = get_very_last(mysql_ptr)))
			goto out;
		printf("Changing attributes of current song %s\n", bn);
		tmp = escape_str(mysql_ptr, bn);
		bn = tmp;
		snprintf(buf2, STRINGSIZE - 1, "update data set %s where "
			"name='%s'", buf, bn);
		ret = update_atts(mysql_ptr, bn, buf2);
//		printf("Updated attributes\n");
	} else { 
		ret = 0;
		while(i < argc && !ret) {
			bn = basename(argv[i]);
			printf("Changing attributes of %s\n", bn);
			snprintf(buf2, STRINGSIZE - 1, 
			"update data set %s where name='%s'", buf, 
				escape_str(mysql_ptr, bn));
			ret = update_atts(mysql_ptr, bn, buf2);
			i++;
		}	
	}
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/* 
 * write stdin to dynamically allocated char array, but maximal max_size byte 
 */
static int stdin2buf(char **buf_ptr, off_t max_size)
{
	const off_t chunk_size = 1024;
	off_t size = 2048;
	char *buf = malloc(size * sizeof(char)), *p = buf;
	int ret;

	if (!buf)
		return -1;
	while ((ret = read(STDIN_FILENO, p, chunk_size)) > 0) {
		p += ret;
		if ((p - buf) + chunk_size > size) {
			char *tmp;

			size *= 2;
			if (size > max_size) {
				ret = -1;
				goto out;
			}
			tmp = realloc(buf, size);
			if (!tmp) {
				ret = -1;
				goto out;
			}
			p = (p - buf) + tmp;
			buf = tmp;
		}
	}
	if (ret < 0)
		goto out;
	*buf_ptr = buf;
	ret = p - buf;
out:
	if (ret < 0 && buf)
		free(buf);
	return ret;

}

/* stradd/pic_add */

int com_stradd_picadd(int argc, char *argv[])
{
	void *mysql_ptr = NULL;
	char *blob = NULL, *esc_blob = NULL, *query = NULL, *fmt;
	int ret;
	off_t size;
	

	if (argc < 3 || strlen(argv[2]) >= MAXLINE - 1)
		return -1;
	if (!strcmp(argv[1], "picadd")) {
		size = MEDIUM_BLOB_SIZE;
		fmt = "INSERT INTO pics (name, pic) values ('%s','%s')";
	} else {
		size = BLOB_SIZE;
		fmt = "INSERT INTO streams (name, def) values ('%s','%s')";
	}

	if ((ret = stdin2buf(&blob, size)) < 0)
		return -1;
	size = ret;
	ret = -1;
	esc_blob = malloc(2 * size);
	query = malloc(2 * size + STRINGSIZE);
	if (!esc_blob || !query)
		goto out;
	if (!(mysql_ptr = init()))
		goto out;
	mysql_real_escape_string(mysql_ptr, esc_blob, blob, size);
	sprintf(query, fmt, argv[2], esc_blob);
	if (mysql_real_query(mysql_ptr, query, strlen(query)))
		goto out;
	ret = 1;
out:
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	if (esc_blob)
		free(esc_blob);
	if (blob)
		free(blob);
	if (query)
		free(query);
	return ret;
}

/* streams */
static int com_streams(int argc, char *argv[])
{
	void *mysql_ptr = NULL;
	void *result = NULL;
	unsigned int num_rows;
	int ret;

	if (!(mysql_ptr = init()))
		return -1;
	result = get_result(mysql_ptr, "select name from streams");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if(!num_rows) {
		ret = 0;
		goto out;
	}
	ret = print_results(result, 0, 0, num_rows, 0);
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

/* strq */
static int com_strq(int argc, char *argv[])
{
	void *mysql_ptr = NULL;
	MYSQL_ROW row;
	char query[STRINGSIZE];
	void *result = NULL;
	unsigned long *length;
	int ret;

	if (argc < 3)
		return -1;
	sprintf(query, "select def from streams where name='%s'", argv[2]);
	if (!(mysql_ptr = init()))
		return -1;
//	printf("q:%s\n", query);
	result = get_result(mysql_ptr, query);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	length = mysql_fetch_lengths(result);
	ret = write(1, row[0], *length);
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

/* strdel */
static int com_strdel(int argc, char *argv[])
{
	void *mysql_ptr = NULL;
	char query[STRINGSIZE];
	int ret = 0;

	if (argc < 3)
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	sprintf(query, "delete from streams where name='%s'", argv[2]);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("stream delete failed\n");
		ret = -1;
		goto out;
	}
out:
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * summary 
 */
static int com_summary(int argc, char *argv[]) 
{ 
	int ret;
	MYSQL_ROW row;
	MYSQL_ROW row2;
	void *mysql_ptr = NULL;
	void *result = NULL;
	void *result2 = NULL;
	char buf[STRINGSIZE];

	if (!(mysql_ptr = init()))
		return -1;
	result = getaa(mysql_ptr);
	if (!result) {
		ret = -1;
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		if (!row[0]) {
			ret = -1;
			goto out;
		}
		sprintf(buf, "select count(name) from data where %s='1'", 
			row[0]);
//		printf("query: %s\n", buf);
		result2 = get_result(mysql_ptr, buf);
		if (!result2) {
			ret = -1;
			goto out;
		}
		row2 = mysql_fetch_row(result2);
		if (!row2 || !row2[0]) {
			ret = -1;
			goto out;
		}
		printf("%s\t%s\n", row[0], row2[0]);
		ret = 0;
	}	
out:	
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * new attribute
 */
static int com_na(int argc, char *argv[]) 
{
	void *mysql_ptr;
	char buf[STRINGSIZE];
	int ret;

	if (argc < 3) {
		printf("na: Syntax error\n");
		return -1;
	}
	if (!(mysql_ptr = init()))
		return -1;
	snprintf(buf, MAXLINE, "alter table data add %s char(1) "
		"NOT NULL default 0", argv[2]);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret) {
		printf("na: alter table failed\n");
		ret = -1;
		goto out;
	}
//	snprintf(buf, MAXLINE, "update data set %s='0'", argv[2]);
//	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
out:	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * delete attribute
 */
static int com_da(int argc, char *argv[]) 
{ 
	void *mysql_ptr;
	char buf[STRINGSIZE];
	int ret;

	if (argc < 3) {
		printf("da: Syntax error\n");
		return -1;
	}
	if (!(mysql_ptr = init()))
		return -1;
	snprintf(buf, MAXLINE, "alter table data drop %s", argv[2]);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret) {
		printf("%s: alter table failed\n", __func__);
		ret = -1;
		goto out;
	}
	ret = 1;
out:	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * verify / clean
 */
static int com_vrfy(int argc, char *argv[]) 
{
	void *mysql_ptr;
	char buf[STRINGSIZE];
	int ret = 0, vrfy_mode = strcmp(argv[1], "clean")? 1 : 0;
	void *result = NULL;
//	unsigned int num_rows, num_nulls = 0;
	unsigned int num_rows;
	MYSQL_ROW row;
	void *result2 = NULL;

	sprintf(buf, "select data.name from data left join dir on "
		"dir.name = data.name where dir.name is NULL ");
	if (!(mysql_ptr = init()))
		return -1;
	result = get_result(mysql_ptr, buf);
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (num_rows) {
		if (vrfy_mode) {
			printf("found %i invalid entr%s\n", num_rows, 
				num_rows == 1? "y" : "ies");
			print_results(result, 0, 0, num_rows - 1, 0);
		} else { 
			MYSQL_ROW row;
			char *escaped_name;
			while ((row = mysql_fetch_row(result))) {
				if (!row[0]) {
					ret = -1;
					goto out;
				}
				escaped_name = escape_str(mysql_ptr,	
					row[0]);
				sprintf(buf, "delete from data where "
					"name = '%s'", escaped_name);
				printf("deleting %s\n", escaped_name);
				ret = mysql_real_query(mysql_ptr, buf, 
					strlen(buf));
				if (ret) {
					printf("failed to delete %s\n", 
						row[0]);
					ret = -1;
					goto out;
				}
			}
		}
	} else
		printf("No invalid entries\n");
	/* look for invalid pic pointers */
	ret = -1;
	if (result)
		mysql_free_result(result);
	result = get_result(mysql_ptr, 
		"select data.name from data left join pics on "
		"pics.id = data.Pic_Id where pics.id is NULL");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (num_rows) {
		if (vrfy_mode) {
			printf("Found %d invalid pic pointer%s:\n", num_rows,
				num_rows == 1? "" : "s");
			print_results(result, 0, 0, num_rows, 0); 
		} else {
			while ((row = mysql_fetch_row(result)) && row[0]) {
//				printf("Fixing %s\n", row[0]);
				sprintf(buf, "update data set Pic_Id=1 "
					"where name = '%s'", 
					escape_str(mysql_ptr, row[0]));
				ret = mysql_real_query(mysql_ptr, buf, 
					strlen(buf));
				if (ret) {
					ret = -1;
					goto out;
				}
			}
			printf("Fixed %d invalid pic pointer%s\n", num_rows,
				num_rows == 1? "" : "s");
		}
	} else
		printf("No invalid pic pointers\n");
	ret = 0;

out:	
	if (result)
		mysql_free_result(result);
	if (result2)
		mysql_free_result(result2);
	if (mysql_ptr)
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * update database
 */
static int com_upd(int argc, char *argv[]) 
{

        char tempfile[_POSIX_PATH_MAX];
        char buf[MAXLINE];
	int ret = 0;
	int num = 0;
	void *mysql_ptr = NULL;
	FILE *pipe = NULL;
	FILE *out_fd = NULL;
	void *result = NULL;
	unsigned int num_rows;
	MYSQL_ROW row;

	if (!(mysql_ptr = init()))
		return -1;
	//printf("find_cmd: xxx%sxxx\n", args_info.find_cmd_arg);
	if (!(pipe = popen(args_info.find_cmd_arg, "r"))) {
		printf("Cannot open pipe\n");
                ret = -1;
		goto out;
	}
	snprintf(tempfile, MAXLINE - 1, "%s.%i", args_info.temp_prefix_arg, 
		getpid());
	if (!(out_fd = fopen(tempfile, "w"))) {
		printf("Cannot open temp file %s\n", tempfile);
                ret = -1;
		goto out;
	}
	while ((fgets(buf, MAXLINE - 1, pipe)) && buf[0]) {
		if (fputs(buf, out_fd) == EOF) {
			printf("Write errror\n");
			ret = -1;
			goto out;
		}
		num++;
	}
	fclose(out_fd);
//	printf("wrote tempfile %s\n", tempfile);
	out_fd = NULL;
	sprintf(buf, "delete from dir"); 
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret) {
		printf("mysql error while flushing table dir. "
			"Return value: %i\n", ret);
		ret = -1;
		goto out;
	}
	sprintf(buf, "load data local infile '%s' into table dir "
		"FIELDS ESCAPED BY \"\"", tempfile);
//	printf("executing mysql query\n%s\n", buf);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret) 
		printf("mysql error. Return value: %i\n", ret);
	else 
		printf("replaced table dir (%i entr%s)\n", num,
			num == 1? "y" : "ies");
	sprintf(buf, "select dir.name from dir left join data on "
		"data.name = dir.name where data.name is NULL");

	result = get_result(mysql_ptr, buf);
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (!num_rows) {
		printf("No new entries\n");
		ret = 0;
		goto out;
	}
	if (!(out_fd = fopen(tempfile, "w"))) {
		printf("Cannot open temp file %s\n", tempfile);
                ret = -1;
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		if (!row[0]) {
			ret = -1;
			goto out;
		}
		printf("New entry: %s\n", row[0]);
		if (fputs(row[0], out_fd) == EOF || fputc('\n', out_fd) == EOF) {
			printf("Write errror\n");
			ret = -1;
			goto out;
		}
	}
	fflush(out_fd);
	sprintf(buf, "load data local infile '%s' into table data "
		"FIELDS ESCAPED BY \"\"", tempfile);
	ret = mysql_real_query(mysql_ptr, buf, strlen(buf));
	if (ret) {
		printf("mysql error. Return value: %i\n", ret);
		goto out;
	}
	printf("found %i new entr%s\n", num_rows, num_rows == 1? "y" : "ies");
	strcpy(buf, "update data set pic_id=1 where pic_id=0");
	mysql_real_query(mysql_ptr, buf, strlen(buf));
	
out:
	if (out_fd)
		fclose(out_fd);
	unlink(tempfile);
	if (pipe)
		pclose(pipe);
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

static long unsigned get_pic_id_by_songname(void *mysql_ptr, char *name)
{
	char query[STRINGSIZE];
	void *result = NULL;
	long unsigned ret;
	MYSQL_ROW row;

	name = escape_str(mysql_ptr, name);
	sprintf(query, "select Pic_Id from data where name = '%s'", name);
	result = get_result(mysql_ptr, query);
	ret = 0;
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = atol(row[0]);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/*
 * pic: get picture by songname
 */
static int com_pic(int argc, char *argv[]) 
{ 
	void *mysql_ptr = NULL;
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length;
	unsigned long id;
	char query[STRINGSIZE];
	int ret;
	char *name = NULL;

	if (!(mysql_ptr = init()))
		return -1;
	if (argc < 3) {
		name = get_very_last(mysql_ptr);
		if (!name)
			goto out;
	}
	else {
		if (strlen(argv[2]) > MAXLINE)
			return -1;
		name = basename(argv[2]);
	}
	if (name[0] == '#') {
		name++;
		id = atoi(name);
	} else
		id = get_pic_id_by_songname(mysql_ptr, name);
	if (!id)
		goto out;
	sprintf(query, "select pic from pics where id='%lu'", id);
//	printf("q:%s\n", query);
	result = get_result(mysql_ptr, query);
	ret = -1;
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	length = mysql_fetch_lengths(result);
	ret = write(1, row[0], *length);
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}
/*
 * picch: change entry's name in pics table
 */
static int com_picch(int argc, char *argv[]) 
{ 
	void *mysql_ptr = NULL;
	int ret;
	long unsigned id;
	char query[STRINGSIZE];

	if (argc != 4)
		return -1;
	id = atol(argv[2]);
	if (!id)
		return -1;
	if (strlen(argv[3]) > MAXLINE)
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	ret = -1;
	sprintf(query, "update pics set name='%s' where id=%lu", argv[3], id);
	if (mysql_real_query(mysql_ptr, query, strlen(query)))
		goto out;
	ret = 0;
out:
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}
/*
 * piclist: print list of pics in db
 */
static int com_piclist(int argc, char *argv[]) 
{
	void *mysql_ptr = NULL;
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length;
	int ret;

	if (!(mysql_ptr = init()))
		return -1;
	ret = -1;
	result = get_result(mysql_ptr, 
		"select id,name,pic from pics order by id");
	if (!result)
		goto out;
	printf("id\tbytes\tname\n");
	while ((row = mysql_fetch_row(result))) {
		length = mysql_fetch_lengths(result);
		if (!row || !row[0] || !row[1] || !row[2])
			goto out;
		printf("%s\t%lu\t%s\n", row[0], length[2], row[1]);
	}
	ret = 0;
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * picass: associate pic to song
 */
static int com_picass(int argc, char *argv[]) 
{
	void *mysql_ptr = NULL;
	void *result = NULL;
//	MYSQL_ROW row;
	char query[STRINGSIZE];
	long unsigned id;
	int i, ret;

	if (argc < 4)
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	id = atol(argv[2]);
	for (i = 3; i < argc; i++) {
		sprintf(query, "update data set Pic_Id=%lu where name='%s'", 
			id, escape_str(mysql_ptr, basename(argv[i])));
		if (mysql_real_query(mysql_ptr, query, strlen(query))) {
			printf("cannot set Pic_Id for %s (%s)\n", argv[i],
				mysql_error(mysql_ptr));
			goto out;
		}
		
	}
	ret = 0;
out:
	if (result)
		mysql_free_result(result);
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
}

/*
 * picdel: delete picture from database
 */
static int com_picdel(int argc, char *argv[]) 
{
	void *mysql_ptr = NULL;
	char query[STRINGSIZE];
	long unsigned id;
	my_ulonglong aff;
	int i, ret = -1;

	if (argc < 3)
		return -1;
	if (!(mysql_ptr = init()))
		return -1;
	for (i = 2; i < argc; i++) {
		if (strlen(argv[i]) > MAXLINE)
			goto out;
		id = atol(argv[i]);
		if (!id)
			goto out;
		sprintf(query, "delete from pics where id=%lu", id);
		if (mysql_real_query(mysql_ptr, query, strlen(query))) {
			printf("cannot delete pic #%lu\n", id);
			goto out;
		}
		aff = mysql_affected_rows(mysql_ptr);
		if (!aff) {
			printf("No such id: %lu\n", id);
			continue;
		}
		if (aff > 1)
			printf("Uuuuuu: deleted %llu pics\n", aff);
		else
			printf("deleted pic #%lu\n", id);
	}
	ret = 0;
out:
	mysql_close(mysql_ptr);
	return ret;
}

static int com_cdb(int argc, char *argv[]) 
{
	void *mysql_ptr = NULL;
	char query[STRINGSIZE];
	int ret = -1;
	char *name;

	if (argc < 3)
		name = "music";
	else {
		name = argv[2];
		if (strlen(name) > MAXLINE)
			goto out;
	}
	/* dont use any database */
	args_info.mysql_database_arg = "";
	if (!(mysql_ptr = init()))
		return -1;
	sprintf(query, "CREATE DATABASE %s", name);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("cannot create database %s\n", name);
		goto out;
	}
	/* reconnect with database just created */
	mysql_close(mysql_ptr);
	args_info.mysql_database_arg = name;
	if (!(mysql_ptr = init()))
		return -1;
	sprintf(query, 
		"create table data (Name VARCHAR(255) BINARY NOT NULL "
		"PRIMARY KEY, "
		"Lastplayed datetime NOT NULL DEFAULT "
			"'1970-01-01', "
		"Numplayed int NOT NULL DEFAULT 0, "
		"Pic_Id bigint unsigned not null DEFAULT 1)"
	);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("cannot create table data\n");
		goto out;
	}
	sprintf(query, 
		"create table dir (Name VARCHAR(255) BINARY NOT NULL "
		"PRIMARY KEY, Dir VARCHAR(255) DEFAULT NULL)"
	);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("cannot create table dir\n");
		goto out;
	}
	sprintf(query, 
		"create table pics ("
		"id bigint(20) unsigned not null primary key "
			"auto_increment, "
		"name varchar(255) binary not null, "
		"pic mediumblob not null)"
	);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("cannot create table pics\n");
		goto out;
	}
	sprintf(query, 
		"create table streams ("
		"name varchar(255) binary not null primary key, "
		"def blob not null)"
	);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		printf("cannot create table streams\n");
		goto out;
	}
	ret = 0;
	printf("Successfully created database %s\n", name);
out:
	if (mysql_ptr);
		mysql_close(mysql_ptr);
	return ret;
	
}

/*
 * Main 
 */
int main(int argc, char *argv[]) 
{
	int i;
	char **p = NULL;
	int ret = -1;

	if (read_config(argc, argv) < 0)
		return -1;
	if (args_info.inputs_num < 1) {
		cmdline_parser_print_help();
		return -1;
	}
	p = malloc((3 + args_info.inputs_num) * sizeof(char *));
	if (!p)
		return -1;
	p[0] = argv[0];
	i = 0;
	while (i < args_info.inputs_num) {
		p[i + 1] = args_info.inputs[i];
		i++;
	}
#if 0
	printf("argc: %d, args_info.inputs_num: %d\n", argc, 
		args_info.inputs_num);
	i = 0;
	while (i < args_info.inputs_num + 1) {
		printf("i:%d, argv: %s, p: %s\n", i, argv[i], p[i]);
		i++;
	}
#endif 
	for (i = 0; cmds[i].func; i++) {
		if (!strcmp(cmds[i].name, p[1])) {
			ret = cmds[i].func(args_info.inputs_num + 1, p);
			goto out;
		}
	}
	printf("Invalid command: \"%s\"\n", args_info.inputs[0]);
out:
	if (p)
		free(p);
	return ret;
}
